@@ -1,16 +1,13 @@
 #!/usr/bin/python3 -B
 
 """
-Easy image file reading & writing. Supports PGM/PPM/PFM/PNG/JPG.
+Easy image file reading & writing. Supports PGM/PPM/PNM/PFM/PNG/JPG/TIFF/EXR/RAW.
 
 Example:
   image, maxval = imgio.imread("foo.png")
   imgio.imwrite("foo.ppm", image, maxval)
 """
 
-from __future__ import absolute_import
-from __future__ import print_function as __print
-
 import os                         # standard library
 import sys                        # standard library
 import unittest                   # standard library
@@ -18,6 +15,7 @@ import unittest                   # standard library
 import piexif                     # pip install piexif
 import numpy as np                # pip install numpy
 import imread as _imread          # pip install imread
+import pyexr                      # pip install pyexr
 
 try:
     # package mode
@@ -35,10 +33,9 @@ except ImportError:
 ######################################################################################
 
 RW_FORMATS = [".pnm", ".pgm", ".ppm", ".pfm", ".png", ".jpg", ".jpeg", ".tif", ".tiff", ".insp", ".raw"]
-RO_FORMATS = RW_FORMATS
+RO_FORMATS = RW_FORMATS + [".exr"]
 
-def imread(filespec, width=None, height=None, bpp=None, verbose=False):
-    # type: (str, int, int, int, bool) -> Tuple[np.ndarray, Union[int, float]]
+def imread(filespec, width=None, height=None, bpp=None, raw_header_size=None, verbose=False):
     """
     Reads the given image file from disk and returns it as a NumPy array.
     Grayscale images are returned as 2D arrays of shape H x W, color images
@@ -56,15 +53,18 @@ def imread(filespec, width=None, height=None, bpp=None, verbose=False):
         _enforce(isinstance(bpp, int) and 1 <= bpp <= 16, "bpp must be an integer in [1, 16]; was %s"%(repr(bpp)))
         _enforce(isinstance(width, int) and width >= 1, "width must be an integer >= 1; was %s"%(repr(width)))
         _enforce(isinstance(height, int) and height >= 1, "height must be an integer >= 1; was %s"%(repr(height)))
-        frame, maxval = _reraise(lambda: _read_raw(filespec, width, height, bpp, verbose=verbose))
+        frame, maxval = _reraise(lambda: _read_raw(filespec, width, height, bpp, raw_header_size, verbose=verbose))
         return frame, maxval
-    elif filetype == ".pfm":
+    if filetype == ".pfm":
         frame, scale = _reraise(lambda: pfm.read(filespec, verbose))
         return frame, scale
-    elif filetype in [".pnm", ".pgm", ".ppm"]:
+    if filetype == ".exr":
+        frame, maxval = _reraise(lambda: _read_exr(filespec, verbose))
+        return frame, maxval
+    if filetype in [".pnm", ".pgm", ".ppm"]:
         frame, maxval = _reraise(lambda: pnm.read(filespec, verbose))
         return frame, maxval
-    elif filetype in [".png", ".tif", ".tiff", ".jpg", ".jpeg", ".insp"]:
+    if filetype in [".png", ".tif", ".tiff", ".jpg", ".jpeg", ".insp"]:
         _print(verbose, "Reading file %s "%(filespec), end='')
         formatstr = "jpg" if filetype == ".insp" else filetype[1:]
         frame = _reraise(lambda: _imread.imread(filespec, formatstr=formatstr))
@@ -76,24 +76,22 @@ def imread(filespec, width=None, height=None, bpp=None, verbose=False):
         c = frame.shape[2] if frame.ndim > 2 else 1
         _print(verbose, "(w=%d, h=%d, c=%d, maxval=%d)"%(w, h, c, maxval))
         return frame, maxval
-    else:
-        raise ImageIOError("unrecognized file type `%s`."%(filetype))
+    raise ImageIOError("unrecognized file type `%s`."%(filetype))
 
 def imwrite(filespec, image, maxval=255, packed=False, verbose=False):
-    # type: (str, np.ndarray, Union[int, float], bool, bool) -> None
     """
-    Writes the given image to the given file. Grayscale images are expected
-    to be provided as NumPy arrays with shape H x W, color images with shape
-    H x W x 3. Metadata, alpha channels, etc. are not supported.
+    Writes the given image to the given file, returns nothing. Grayscale images
+    are expected to be provided as NumPy arrays with shape H x W, color images
+    with shape H x W x 3. Metadata, alpha channels, etc. are not supported.
     """
     ImageIOError.error_message_prefix = "Failed to write %s: "%(repr(filespec))
     _enforce(isinstance(filespec, str), "filespec must be a string, was %s (%s)."%(type(filespec), repr(filespec)))
     _enforce(isinstance(image, np.ndarray), "image must be a NumPy ndarray; was %s."%(type(image)))
-    _enforce(image.dtype in [np.uint8, np.uint16, np.float32], "image.dtype must be uint8, uint16, or float32; was %s"%(image.dtype))
+    _enforce(image.dtype.char in "BHefd", "image.dtype must be uint{8,16}, or float{16,32,64}; was %s"%(image.dtype))
     _enforce(image.size >= 1, "image must have at least one pixel; had none.")
     _enforce(isinstance(maxval, (float, int)), "maxval must be an integer or a float; was %s."%(repr(maxval)))
-    _enforce(isinstance(maxval, int) or image.dtype == np.float32, "maxval must be an integer in [1, 65535]; was %s."%(repr(maxval)))
-    _enforce(1 <= maxval <= 65535 or image.dtype == np.float32, "maxval must be an integer in [1, 65535]; was %s."%(repr(maxval)))
+    _enforce(isinstance(maxval, int) or image.dtype.char in "efd", "maxval must be an integer in [1, 65535]; was %s."%(repr(maxval)))
+    _enforce(1 <= maxval <= 65535 or image.dtype.char in "efd", "maxval must be an integer in [1, 65535]; was %s."%(repr(maxval)))
     _enforce(isinstance(verbose, bool), "verbose must be True or False, was %s (%s)."%(type(verbose), repr(verbose)))
     _disallow(image.ndim not in [2, 3], "image.shape must be (m, n) or (m, n, 3); was %s."%(str(image.shape)))
     _disallow(image.ndim == 3 and image.shape[2] != 3, "image.shape must be (m, n) or (m, n, 3); was %s."%(str(image.shape)))
@@ -110,7 +108,7 @@ def imwrite(filespec, image, maxval=255, packed=False, verbose=False):
         _reraise(lambda: _write_raw(filespec, image, maxval, packed, verbose))
     elif filetype == ".pfm":
         _enforce(maxval >= 0.0, "maxval (scale) must be non-negative; was %s."%(repr(maxval)))
-        _reraise(lambda: pfm.write(filespec, image, maxval, verbose))
+        _reraise(lambda: pfm.write(filespec, image, maxval, little_endian=True, verbose=verbose))
     elif filetype in [".pnm", ".pgm", ".ppm"]:
         _reraise(lambda: pnm.write(filespec, image, maxval, verbose))
     elif filetype in [".png", ".tif", ".tiff", ".jpg", ".jpeg", ".insp"]:
@@ -147,8 +145,6 @@ class ImageIOError(RuntimeError):
 #
 ######################################################################################
 
-# pylint: disable=missing-docstring
-
 def _enforce(expression, error_message_if_false):
     if not expression:
         raise ImageIOError("%s"%(error_message_if_false))
@@ -160,8 +156,8 @@ def _disallow(expression, error_message_if_true):
 def _reraise(func):
     try:
         return func()
-    except Exception:
-        raise ImageIOError("%s"%(repr(sys.exc_info()[1])))
+    except Exception as e:
+        raise ImageIOError("%s"%(repr(sys.exc_info()[1]))) from e
 
 def _print(verbose, *args, **kwargs):
     if verbose:
@@ -173,20 +169,27 @@ def _exif_rotate(img, filespec):
         exif_dict = piexif.load(filespec).pop("0th")
         exif_orientation = exif_dict.get(piexif.ImageIFD.Orientation)
         orientation = 1 if exif_orientation is None else exif_orientation
-    except piexif._exceptions.InvalidImageDataError as e:
+    except piexif.InvalidImageDataError:
         pass
-    finally:
-        exif_to_rot90 = {1:0, 2:0, 3:2, 4:0, 5:1, 6:3, 7:3, 8:1}
-        if orientation in [2, 5, 7]:
-            img = np.fliplr(img)
-        if orientation in [4]:
-            img = np.flipud(img)
-        if orientation in exif_to_rot90:
-            rot90_ccw_steps = exif_to_rot90[orientation]
-            img = np.rot90(img, rot90_ccw_steps)  # 0/90/180/270 CCW
-        return img
+    exif_to_rot90 = {1: 0, 2: 0, 3: 2, 4: 0, 5: 1, 6: 3, 7: 3, 8: 1}
+    if orientation in [2, 5, 7]:
+        img = np.fliplr(img)
+    if orientation in [4]:
+        img = np.flipud(img)
+    if orientation in exif_to_rot90:
+        rot90_ccw_steps = exif_to_rot90[orientation]
+        img = np.rot90(img, rot90_ccw_steps)  # 0/90/180/270 CCW
+    return img
 
-def _read_raw(filespec, width, height, bpp, verbose=False):
+def _read_exr(filespec, verbose=True):
+    exr = pyexr.open(filespec)
+    data = exr.get()
+    maxval = np.max(data)
+    _print(verbose, "Reading OpenEXR file %s "%(filespec), end='')
+    _print(verbose, "(w=%d, h=%d, c=%d, %s)"%(exr.width, exr.height, len(exr.channels), data.dtype))
+    return data, maxval
+
+def _read_raw(filespec, width, height, bpp, header_size=None, verbose=False):
     # Warning: hardcoded endianness (x86)
     with open(filespec, "rb") as infile:
         buf = infile.read()
@@ -194,20 +197,22 @@ def _read_raw(filespec, width, height, bpp, verbose=False):
         maxval = 2 ** bpp - 1
         wordsize = 2 if bpp > 8 else 1
         packed = len(buf) < (width * height * wordsize)
+        if header_size is None:
+            header_size = len(buf) - (width * height * wordsize)
         _print(verbose, "Reading raw Bayer file %s "%(filespec), end='')
-        _print(verbose, "(w=%d, h=%d, maxval=%d, packed=%r)"%(width, height, maxval, packed))
+        _print(verbose, "(w=%d, h=%d, maxval=%d, header=%d, packed=%r)"%(width, height, maxval, header_size, packed))
         if not packed:
             dtype = "<u2" if bpp > 8 else np.uint8
-            pixels = np.frombuffer(buf, dtype, count=width * height, offset=0)
+            pixels = np.frombuffer(buf, dtype, count=width * height, offset=header_size)
             pixels = pixels.reshape(shape).astype(np.uint8 if bpp <= 8 else np.uint16)
         else:
-            # TODO: unpack!
+            # TODO: packed raw support
             raise ImageIOError("Packed RAW reading not implemented yet!")
         return pixels, maxval
 
-def _write_raw(filespec, image, maxval, pack=False, verbose=False):
+def _write_raw(filespec, image, _maxval, _pack=False, _verbose=False):
+    # TODO: packed raw support
     # Warning: hardcoded endianness (x86)
-    bpp = int(np.log2(maxval + 1))
     with open(filespec, "wb") as outfile:
         image = image.copy(order='C')  # ensure x86 byte order
         outfile.write(image)
@@ -233,7 +238,7 @@ class _TestImgIo(unittest.TestCase):
             expected_name = expected_exception.__name___
             errstr = "Expected %s with a message matching '%s', got %s."%(expected_name, expected_regex, raised_name)
             print("   FAIL: %s"%(errstr))
-            raise AssertionError(errstr)
+            raise AssertionError(errstr) from e
         try:  # then check the exception message
             assertRaisesRegex = getattr(unittest.TestCase, "assertRaisesRegex", unittest.TestCase.assertRaisesRegexp)
             assertRaisesRegex(self, expected_exception, expected_regex, *args, **kwargs)  # python 2 vs. 3 compatibility
@@ -248,7 +253,7 @@ class _TestImgIo(unittest.TestCase):
             raise
 
     def test_exceptions(self):
-        # pylint: disable=deprecated-method
+        # pylint: disable=too-many-statements
         print("Testing exception handling...")
         shape = (7, 11, 3)
         pixels = np.random.random(shape).astype(np.float32)
@@ -301,7 +306,7 @@ class _TestImgIo(unittest.TestCase):
         self.assertRaisesRegexp(ImageIOError, "^Failed to write", imwrite, "imgio.test.ppm", np.zeros((7, 7, 2), np.uint8))
         self.assertRaisesRegexp(ImageIOError, "^Failed to write", imwrite, "imgio.test.ppm", np.zeros((7, 7, 4), np.uint8))
         self.assertRaisesRegexp(ImageIOError, "^Failed to write", imwrite, "imgio.test.ppm", np.zeros((7, 7, 3, 1), np.uint8))
-        self.assertRaisesRegexp(ImageIOError, "^Failed to write", imwrite, "imgio.test.ppm", pixels.astype(np.bool))
+        self.assertRaisesRegexp(ImageIOError, "^Failed to write", imwrite, "imgio.test.ppm", pixels.astype(bool))
         self.assertRaisesRegexp(ImageIOError, "^Failed to write", imwrite, "imgio.test.ppm", pixels.astype(np.float16))
         self.assertRaisesRegexp(ImageIOError, "^Failed to write", imwrite, "imgio.test.ppm", pixels.astype(np.float64))
         self.assertRaisesRegexp(ImageIOError, "^Failed to write", imwrite, "imgio.test.ppm", pixels.astype('>f4'))
@@ -373,7 +378,7 @@ class _TestImgIo(unittest.TestCase):
             pixels = (pixels * 127).astype(np.uint8)
             for extension in ["jpg", "insp"]:
                 tempfile = "imgio.test." + extension
-                print("Testing %s reading & writing..."%(extension.upper()))
+                print("Testing %s reading & writing, shape=%s..."%(extension.upper(), repr(shape)))
                 imwrite(tempfile, pixels, maxval, verbose=False)
                 result, resmaxval = imread(tempfile, verbose=False)
                 self.assertEqual(resmaxval, maxval)
@@ -395,7 +400,7 @@ class _TestImgIo(unittest.TestCase):
             self.assertEqual(resscale, scale)
             self.assertEqual(result.dtype, np.float32)
             self.assertEqual(result.shape, shape)
-            self.assertEqual(result.shape, shape)
+            self.assertTrue(np.allclose(result, pixels))
             os.remove(tempfile)
 
     def test_exif(self):
@@ -415,6 +420,15 @@ class _TestImgIo(unittest.TestCase):
                 self.assertEqual(refimg.shape, testimg.shape)
                 self.assertGreater(np.sum(epsdiff), 0.5 * epsdiff.size)
 
+    def test_exr(self):
+        print("Testing EXR reading...")
+        thispath = os.path.dirname(os.path.abspath(__file__))
+        filespec = os.path.join(thispath, "test-images", "GrayRampsDiagonal.exr")
+        img, maxval = imread(filespec)
+        self.assertEqual(img.shape, (800, 800, 1))
+        self.assertEqual(img.dtype, np.float32)
+        self.assertEqual(maxval, np.max(img))
+
     def test_raw(self):
         for packed in [False]:
             for shape in [(1, 1), (7, 11)]:
@@ -434,6 +448,25 @@ class _TestImgIo(unittest.TestCase):
                     self.assertEqual(result.tolist(), pixels.tolist())
                     os.remove(tempfile)
 
+    def test_raw_header(self):
+        bpp = 12
+        shape = (28, 60)
+        maxval = 2**bpp - 1
+        tempfile = "imgio.test%db.raw"%(bpp)
+        print("Testing RAW header reading & writing...")
+        header = np.arange(17).astype(np.uint16)
+        footer = np.arange(19).astype(np.uint16)
+        pixels = np.random.random(shape)
+        pixels = (pixels * maxval).astype(np.uint16)
+        data = np.hstack((header, pixels.flatten(), footer)).reshape(1, -1)
+        imwrite(tempfile, data, maxval)
+        result, resmaxval = imread(tempfile, width=shape[1], height=shape[0], bpp=bpp, raw_header_size=17 * 2)
+        self.assertEqual(resmaxval, maxval)
+        self.assertEqual(result.dtype, np.uint16)
+        self.assertEqual(result.shape, shape)
+        self.assertEqual(result.tolist(), pixels.tolist())
+        os.remove(tempfile)
+
     def test_allcaps(self):
         print("Testing Windows-style all-caps filenames...")
         maxval = 255
